%% MEDSEM - Multienergy deconvolution, parallel version
%
% Usage - run to select directory where a series of pre-aligned
% SEM images, named "aligned0001.tif" ... "alignedXXXX.tif" are located
% Naming convention (among other options) can be changed below.
% Saves a series of deconvolved images as ".\results\slice01.tif" etc. for
% 3D viewing, or other processing.
%
% X.Zhuge, F. Boughorbel, FEI
% E.Johlin, AMOLF
% Last modified October, 2016
%% Arguments for parallelization
clear all; close all; clc;% delete(gcp);
loadnamepre = 'aligned'; %name prefix; e.g. files are named e.g. "aligned0001.tif"
NsubsampleX = 2;
NsubsampleY = 2;
Nsubsample = NsubsampleX*NsubsampleY;
parpool('local', Nsubsample,'AttachedFiles',{'TVE3D.m'});
%Nworkers = parpool('size'); %use NSection workers
%% File intput
loadpath = uigetdir; % Get input files location
if loadpath == 0 % If the input dialogue is canceled, exit.
    delete(gcp);
    disp('Canceled, exiting...');
    return;
end
LocalSavepath = [loadpath '/result']; %Save files in "./result"
mkdir(LocalSavepath);
%% Parameter settings for deconvolution
ovsmp=2; %determines number of slices to create from each energy BSE image
regularization = 1; % choose whether to regularize the algorithm (TV)
bcadj = 1; % turns brightness/contrast autoadjust (maximization) on/off
% Stop criteria:
costlimit = 3.0e-4; % Convergence at 3.0e-4;
NiterMax = 3000;   % Limit to iterations (3000)
StopMeasureTarget = 1.0e-4; % Limit change in brightness
%% Loading and preprocessing
loadcheck=2;
ii=1;
loadname=strcat(loadnamepre, sprintf('%04d',ii),'.tif'); %load file name, add zeroes to keep 4 digits
while loadcheck==2
    ii=ii+1;
    loadname=strcat(loadnamepre, sprintf('%04d',ii),'.tif');
    loadcheck=exist(strcat(loadpath,'/',loadname), 'file');
end;
if ii <= 2 % If the input dialogue is canceled, exit.
    delete(gcp);
    disp('No images found, exiting...');
    return;
end
Nenergies=ii-1;
Nz=Nenergies*ovsmp;
%% Begin Parallel Session
spmd
    WorkerIndex = labindex();
    Y = [];
    disp('Loading images...');
    for ii=1:Nenergies
        loadname=strcat(loadnamepre, sprintf('%04d',ii),'.tif');
        Y(ii,:,:) = imread([loadpath,'/',loadname]);
    end;
    if bcadj == 1
      Y = double(Y)/(2^16);
      for ii=1:Nenergies
        Y(ii,:,:) = mat2gray(Y(ii,:,:)); %Maximize contrast/frame
      end;
    else
        Y = mat2gray(Y); %Maximize contrast/stack
    end
    %% Setup constants for deconvolution
    [n,ht,wd] = size(Y);

    Npixel = ht*wd; NpixelY = Npixel; htwdY = [ht,wd];

    rowIndex = floor((WorkerIndex-1)/NsubsampleX) + 1;
    colIndex = (WorkerIndex-1) - ((rowIndex-1)*NsubsampleX) + 1;

    Y = Y(:,rowIndex:NsubsampleX:end, colIndex:NsubsampleY:end);

    [n,ht0,wd0] = size(Y);
    [n,Npixel] = size(Y);

    Y = reshape(Y,[Nenergies,Npixel]);

    [n,ht,wd] = size(Y);

    %% Initialization
    sigma_z = linspace(Nz/5,Nz/3,Nenergies);
    zd = linspace(1,Nz,Nenergies);
    z = linspace(1,Nz,Nz);
    Hx = zeros(Nenergies,Nz);
    for ne = 1:Nenergies
        Hx(ne,:) = exp(-(z-zd(ne)).^2./(2*sigma_z(ne)^2));
    end;
    Ox = 0.1*ones(Nz,Npixel);
    %%
    cost = norm(Y-Hx*Ox, 'fro')/Npixel;
    costrecord = [];
    disp('Begin deconvolution');
    
    %Deconvolution settings
    lamda =  10^(-2.5); %determines smoothness of regularizaiton (smaller=smoother) [10^(-2.5)]
    omega = 1.9; %acceleration exponent for imrpoving convergence [1.9]
    alpha = 2; %relaxation exponent for improving stability [2]
    for iter=1:NiterMax

        Oxpre = Ox;

        if regularization == 1
            OxR= (Ox./(1-lamda*TVE3D(Ox,Nz,ht,wd))); %Regularize (prevents noise and distorion for large iterations)
        else
            OxR=Ox; %Don't Regularize
        end;
        Ox = (OxR.*(Hx'*(Y./(Hx*Ox + eps)))); % SI Eq. 8
        Hx = (Hx.*(((Y+eps)./(Hx*Ox + eps)).^alpha * Ox').^(omega/alpha)); % SI Eq. 9
        Hx = bsxfun(@rdivide,Hx,sum(Hx,1));

        cost = norm(Y-Hx*Ox, 'fro')/Npixel; % Calculate auto-shutoff parameters
       % costrecord = [costrecord;cost];
        
        difimage = abs(Ox - Oxpre);
        StopMeasure = sum(difimage(:))/sum(Oxpre(:));
        
        if cost<=costlimit || StopMeasure<StopMeasureTarget % Check auto-shutoff
            disp(['Stop criteria reached at iteration:',int2str(iter),', at cost level:',num2str(cost),', Difference measure:', num2str(StopMeasure)]);
            break;
        end;

        disp(['Iteration:', int2str(iter),', Cost:',num2str(cost),', Difference measure:', num2str(StopMeasure)]); % Display progress
              
    end;
    % Save temp subslice
    namepre = ['subslice', num2str(rowIndex),'_',num2str(colIndex)];
    for k=1:Nz
        img_data = Ox(k,:);
        slice = reshape(img_data, ht0, wd0);
        slice = 1-mat2gray(slice);
     
        savename = strcat(namepre, sprintf('%02d',k),'.tif'); %save temp file name, add zeroes to keep 2 digits
        imwrite((slice), [LocalSavepath,'\',savename]);
    end;

end;

% Save deconvolution weightings 
B=Hx{1};
BB=Hx{2};
csvwrite([LocalSavepath,'\','mixing1.dat'],B)
csvwrite([LocalSavepath,'\','mixing2.dat'],BB)
%% End paralell session
delete(gcp);

% loading the sub-images and re-assembling the full resolution slices 
loadname = strcat(loadnamepre,'0001','.tif');
Y = imread([loadpath,'\',loadname]);
savenamepre = 'slice';
for k=1:Nz
    for ii = 1:Nsubsample
        rowIndex = floor((ii-1)/NsubsampleX) + 1;
        colIndex = (ii-1) - ((rowIndex-1)*NsubsampleX) + 1;

        loadnamepre = ['subslice', num2str(rowIndex),'_',num2str(colIndex)];
        loadname = strcat(loadnamepre, sprintf('%02d',k),'.tif'); 
        Y(rowIndex:NsubsampleX:end, colIndex:NsubsampleY:end)= imread([LocalSavepath,'\',loadname]);
        delete([LocalSavepath,'\',loadname]); %delete temporary images after loading
    end;
    savename = strcat(savenamepre, sprintf('%02d',k),'.tif'); %Save images as e.g. "slice01.tif" 
    full_slice = (mat2gray(Y));
    imwrite((full_slice), [LocalSavepath,'\',savename]); % Save
end;
